# coding: utf-8

"""
Remote filesystem and targets, using a configurable remote file interface for atomic operations.
"""

__all__ = ["RemoteFileSystem", "RemoteTarget", "RemoteFileTarget", "RemoteDirectoryTarget"]


import os
import time
import fnmatch
from contextlib import contextmanager

import six

from law.config import Config
from law.target.file import (
    FileSystem, FileSystemTarget, FileSystemFileTarget, FileSystemDirectoryTarget, get_path,
    get_scheme, add_scheme, remove_scheme,
)
from law.target.local import LocalFileSystem, LocalFileTarget, LocalDirectoryTarget
from law.target.remote.cache import RemoteCache
from law.util import make_list, merge_dicts
from law.logger import get_logger


logger = get_logger(__name__)

_local_fs = LocalFileSystem.default_instance


class RemoteFileSystem(FileSystem):

    default_instance = None
    file_interface_cls = None
    local_fs = _local_fs
    _updated_sections = set()

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        config = super(RemoteFileSystem, cls).parse_config(section, config=config,
            overwrite=overwrite)

        cfg = Config.instance()

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func):
            if option not in config or overwrite:
                config[option] = func(section, option)

        # default setting for validation for existence after copy
        add("validate_copy", cfg.get_expanded_bool)

        # default setting for using the cache
        add("use_cache", cfg.get_expanded_bool)

        # cache options
        if cfg.options(section, prefix="cache_"):
            RemoteCache.parse_config(section, config.setdefault("cache_config", {}),
                overwrite=overwrite)

        return config

    @classmethod
    def _update_section_defaults(cls, default_section, section):
        # do not update the section when it is the default one or it was already updated
        if section == default_section or section in cls._updated_sections:
            return
        cls._updated_sections.add(section)

        # get the defaults
        cfg = Config.instance()
        defaults = dict(cfg.items(default_section, expand_vars=False, expand_user=False))

        # update
        cfg.update({section: defaults}, overwrite_sections=True, overwrite_options=False)

    @classmethod
    def split_remote_kwargs(cls, kwargs, include=None, skip=None):
        """
        Takes keyword arguments *kwargs*, splits them into two separate dictionaries depending on
        their content, and returns them in a tuple. The first one will contain arguments related to
        potential remote file operations (e.g. ``"cache"`` or ``"retries"``), while the second one
        will contain all remaining arguments. This function is used internally to decide which
        arguments to pass to target formatters. *include* (*skip*) can be a list of argument keys
        that are considered as well (ignored).
        """
        include = make_list(include) if include else []
        skip = make_list(skip) if skip else []
        transfer_kwargs = {
            name: kwargs.pop(name)
            for name in ["cache", "prefer_cache", "retries", "retry_delay"] + include
            if name in kwargs and name not in skip
        }
        return transfer_kwargs, kwargs

    def __init__(self, file_interface, validate_copy=False, use_cache=False, cache_config=None,
            local_fs=None, **kwargs):
        super(RemoteFileSystem, self).__init__(**kwargs)

        # store the file interface
        self.file_interface = file_interface

        # store other configs
        self.validate_copy = validate_copy
        self.use_cache = use_cache

        # set the cache when a cache root is set in the cache_config
        if cache_config and cache_config.get("root"):
            self.cache = RemoteCache(self, **cache_config)
        else:
            self.cache = None

        # when passed, store a custom local fs on instance level
        # otherwise, the class level member is used
        if local_fs:
            self.local_fs = local_fs

    def __del__(self):
        # cleanup the cache
        if getattr(self, "cache", None):
            del self.cache
            self.cache = None

    def __repr__(self):
        return "{}({}, name={}, base={}, {})".format(self.__class__.__name__,
            self.file_interface.__class__.__name__, self.name, self.base[0], hex(id(self)))

    def _init_configs(self, section, default_fs_option, default_section, init_kwargs):
        cfg = Config.instance()

        # get the proper section
        if not section:
            section = cfg.get_expanded("target", default_fs_option)

        # try to read it and fill configs to pass to the file system and the remote file interface
        fs_config = {}
        fi_config = {}
        if isinstance(section, six.string_types):
            # when set, the section must exist
            if not cfg.has_section(section):
                raise Exception("law config has no section '{}' to read {} options".format(
                    section, self.__class__.__name__))

            # extend options of sections other than the default one with its values
            self._update_section_defaults(default_section, section)

            # read the configs from the section for both the file system and remote interface
            fs_config = self.parse_config(section)
            fi_config = self.file_interface_cls.parse_config(section)

        # update both configs with init kwargs
        fs_config = merge_dicts(fs_config, init_kwargs, deep=True)
        fi_config = merge_dicts(fi_config, init_kwargs, deep=True)

        return section, fs_config, fi_config

    @property
    def base(self):
        return self.file_interface.base

    def is_local(self, path):
        return get_scheme(path) == "file"

    def abspath(self, path):
        # due to the dynamic definition of remote bases, path is supposed to be already absolute,
        # so just handle leading and trailing slashes when there is no scheme scheme
        path = str(path)
        return ("/" + path.strip("/")) if not get_scheme(path) else path

    def uri(self, path, **kwargs):
        return self.file_interface.uri(self.abspath(path), **kwargs)

    def dirname(self, path):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.dirname(path)

        return super(RemoteFileSystem, self).dirname(self.abspath(path))

    def basename(self, path):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.basename(path)

        return super(RemoteFileSystem, self).basename(self.abspath(path))

    def stat(self, path, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.stat(path)

        return self.file_interface.stat(self.abspath(path), **kwargs)

    def exists(self, path, stat=False, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.exists(path, stat=stat)

        return self.file_interface.exists(self.abspath(path), stat=stat, **kwargs)

    def isdir(self, path, rstat=None, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.isdir(path)

        return self.file_interface.isdir(path, stat=rstat, **kwargs)

    def isfile(self, path, rstat=None, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.isfile(path)

        return self.file_interface.isfile(path, stat=rstat, **kwargs)

    def chmod(self, path, perm, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.chmod(path, perm)

        if not self.has_permissions:
            return True

        return self.file_interface.chmod(self.abspath(path), perm, **kwargs)

    def remove(self, path, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.remove(path)

        # protection against removing the base directory of the remote file system
        path = self.abspath(path)
        if path == "/":
            logger.warning("refused request to remove base directory of {!r}".format(self))
            return

        return self.file_interface.remove(path, **kwargs)

    def mkdir(self, path, perm=None, recursive=True, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            return self.local_fs.mkdir(path, perm=perm, recursive=recursive)

        if perm is None:
            perm = self.default_dir_perm or 0o0770

        func = self.file_interface.mkdir_rec if recursive else self.file_interface.mkdir
        x = func(self.abspath(path), perm, **kwargs)
        return x

    def listdir(self, path, pattern=None, type=None, **kwargs):
        # forward to local_fs
        path = str(path)
        if self.is_local(path):
            return self.local_fs.listdir(path, pattern=pattern, type=type)

        elems = self.file_interface.listdir(self.abspath(path), **kwargs)

        # apply pattern filter
        if pattern is not None:
            elems = fnmatch.filter(elems, pattern)

        # apply type filter
        if type == "f":
            elems = [e for e in elems if not self.isdir(os.path.join(path, e), **kwargs)]
        elif type == "d":
            elems = [e for e in elems if self.isdir(os.path.join(path, e, **kwargs))]

        return elems

    def walk(self, path, max_depth=-1, **kwargs):
        # forward to local_fs
        if self.is_local(path):
            for obj in self.local_fs.walk(path, max_depth=max_depth):
                yield obj
            return

        # mimic os.walk with a max_depth and yield the current depth
        search_dirs = [(self.abspath(path), 0)]
        while search_dirs:
            (search_dir, depth) = search_dirs.pop(0)

            # check depth
            if max_depth >= 0 and depth > max_depth:
                continue

            # find dirs and files
            dirs = []
            files = []
            for elem in self.listdir(search_dir, **kwargs):
                if self.isdir(os.path.join(search_dir, elem), **kwargs):
                    dirs.append(elem)
                else:
                    files.append(elem)

            # yield everything
            yield (search_dir, dirs, files, depth)

            # use dirs to update search dirs
            search_dirs.extend((os.path.join(search_dir, d), depth + 1) for d in dirs)

    def glob(self, pattern, cwd=None, **kwargs):
        # forward to local_fs
        pattern = str(pattern)
        if self.is_local(pattern):
            return self.local_fs.glob(pattern, cwd=cwd)

        # helper to check if a string represents a pattern
        def is_pattern(s):
            return "*" in s or "?" in s

        # prepare pattern
        if cwd is not None:
            pattern = os.path.join(cwd, pattern)

        # split the pattern to determine the search path, i.e. the leading part that does not
        # contain any glob chars, e.g. "foo/bar/test*/baz*" -> "foo/bar"
        search_dir = []
        patterns = []
        for part in pattern.split("/"):
            if not patterns and not is_pattern(part):
                search_dir.append(part)
            else:
                patterns.append(part)
        search_dir = self.abspath("/".join(search_dir))

        # walk trough the search path and use fnmatch for comparison
        elems = []
        max_depth = len(patterns) - 1
        for root, dirs, files, depth in self.walk(search_dir, max_depth=max_depth, **kwargs):
            # get the current pattern
            pattern = patterns[depth]

            # when we are still below the max depth, filter dirs
            # otherwise, filter files and dirs to select
            if depth < max_depth:
                dirs[:] = fnmatch.filter(dirs, pattern)
            else:
                elems += [os.path.join(root, e) for e in fnmatch.filter(dirs + files, pattern)]

        # cut the cwd if there was any
        if cwd is not None:
            elems = [os.path.relpath(e, cwd) for e in elems]

        return elems

    # atomic copy
    def _atomic_copy(self, src, dst, perm=None, validate=None, **kwargs):
        if validate is None:
            validate = self.validate_copy

        src = self.abspath(src)
        dst = self.abspath(dst)

        # actual copy
        src_uri, dst_uri = self.file_interface.filecopy(src, dst, **kwargs)

        # copy validation
        dst_fs = self.local_fs if self.is_local(dst_uri) else self
        if validate:
            if not dst_fs.exists(dst):
                raise Exception("validation failed after copying {} to {}".format(src_uri, dst_uri))

        # handle permissions
        if perm is None:
            perm = dst_fs.default_file_perm
        dst_fs.chmod(dst, perm)

        return dst_uri

    # generic copy with caching ability (local paths must have a "file://" scheme)
    def _cached_copy(self, src, dst, perm=None, cache=None, prefer_cache=False, validate=None,
            **kwargs):
        """
        When this method is called, both *src* and *dst* should refer to files.
        """
        if self.cache is None:
            cache = False
        elif cache is None:
            cache = self.use_cache
        else:
            cache = bool(cache)

        # ensure absolute paths
        src = self.abspath(src)
        dst = dst and self.abspath(dst) or None

        # determine the copy mode for code readability
        # (remote-remote: "rr", remote-local: "rl", remote-cache: "rc", ...)
        src_local = self.is_local(src)
        dst_local = dst and self.is_local(dst)
        mode = "rl"[src_local] + ("rl"[dst_local] if dst is not None else "c")

        # disable caching when the mode is local-local, local-cache or remote-remote
        if mode in ("ll", "lc", "rr"):
            cache = False

        # dst can be None, but in this case, caching should be enabled
        if dst is None and not cache:
            raise Exception("copy destination must not be empty when caching is disabled")

        if not cache:
            # simply copy and return the dst path
            return self._atomic_copy(src, dst, perm=perm, validate=validate, **kwargs)

        kwargs_no_retries = kwargs.copy()
        kwargs_no_retries["retries"] = 0

        # handle 3 cases: lr, rl, rc
        if mode == "lr":
            # strategy: copy to remote, copy to cache, sync stats

            # copy to remote, no need to validate as we compute the stat anyway
            dst_uri = self._atomic_copy(src, dst, perm=perm, validate=False, **kwargs)
            rstat = self.stat(dst, **kwargs_no_retries)

            # remove the cache entry
            if dst in self.cache:
                logger.debug("removing destination file {} from cache".format(dst))
                self.cache.remove(dst)

            # allocate cache space and copy to cache
            lstat = self.local_fs.stat(src)
            self.cache.allocate(lstat.st_size)
            cdst_uri = add_scheme(self.cache.cache_path(dst), "file")
            with self.cache.lock(dst):
                logger.debug("loading source file {} to cache".format(src))
                self._atomic_copy(src, cdst_uri, validate=False)
                self.cache.touch(dst, (int(time.time()), rstat.st_mtime))

            return dst_uri

        else:  # rl, rc
            # strategy: copy to cache when not up to date, sync stats, opt. copy to local

            # build the uri to the cache path of the src file
            csrc_uri = add_scheme(self.cache.cache_path(src), "file")

            # if the file is cached and prefer_cache is true,
            # return the cache path, no questions asked
            # otherwise, check if the file is there and up to date
            if not prefer_cache or src not in self.cache:
                with self.cache.lock(src):
                    # in cache and outdated?
                    rstat = self.stat(src, **kwargs_no_retries)
                    if src in self.cache and not self.cache.check_mtime(src, rstat.st_mtime):
                        logger.debug("source file {} is outdated in cache, removing".format(src))
                        self.cache.remove(src, lock=False)
                    # in cache at all?
                    if src not in self.cache:
                        self.cache.allocate(rstat.st_size)
                        self._atomic_copy(src, csrc_uri, validate=validate, **kwargs)
                        logger.debug("loading source file {} to cache".format(src))
                        self.cache.touch(src, (int(time.time()), rstat.st_mtime))

            if mode == "rl":
                # simply use the local_fs for copying
                self.local_fs.copy(csrc_uri, dst, perm=perm)
                return dst

            # mode is rc
            return csrc_uri

    def _prepare_dst_dir(self, dst, src=None, perm=None, **kwargs):
        """
        Prepares the directory of a target located at *dst* for copying and returns its full
        location as specified below. *src* can be the location of a source file target, which is
        (e.g.) used by a file copy or move operation. When *dst* is already a directory, calling
        this method has no effect and the *dst* path is returned, optionally joined with the
        basename of *src*. When *dst* is a file, *dst* path is returned unchanged. Otherwise, when
        *dst* does not exist yet, it is interpreted as a file path and missing directories are
        created when :py:attr:`create_file_dir` is *True*, using *perm* to set the directory
        permission. *dst* is returned.
        """
        rstat = self.exists(dst, stat=True)

        if rstat:
            if self.file_interface.isdir(dst, stat=rstat) and src:
                full_dst = os.path.join(dst, os.path.basename(src))
            else:
                full_dst = dst

        else:
            # interpret dst as a file name, create missing dirs
            dst_dir = self.dirname(dst)
            if dst_dir and self.create_file_dir and not self.isdir(dst_dir):
                self.mkdir(dst_dir, perm=perm, recursive=True, **kwargs)
            full_dst = dst

        return full_dst

    def copy(self, src, dst, perm=None, dir_perm=None, **kwargs):
        # dst might be an existing directory
        if dst:
            dst_fs = self.local_fs if self.is_local(dst) else self
            dst_fs._prepare_dst_dir(dst, src=src, perm=dir_perm, **kwargs)

        # copy the file
        return self._cached_copy(src, dst, perm=perm, **kwargs)

    def move(self, src, dst, perm=None, dir_perm=None, **kwargs):
        if not dst:
            raise Exception("move requires dst to be set")

        # copy the file
        kwargs["cache"] = False
        kwargs.setdefault("validate", True)
        dst = self.copy(src, dst, perm=perm, dir_perm=dir_perm, **kwargs)

        # remove the src
        src_fs = self.local_fs if self.is_local(src) else self
        src_fs.remove(src, **kwargs)

        return dst

    def open(self, path, mode, perm=None, dir_perm=None, cache=None, **kwargs):
        if self.cache is None:
            cache = False
        elif cache is None:
            cache = self.use_cache
        else:
            cache = bool(cache)

        yield_path = kwargs.pop("_yield_path", False)
        path = self.abspath(path)
        tmp = None
        read_mode = mode.startswith("r")

        if read_mode:
            if cache:
                lpath = self._cached_copy(path, None, cache=cache, **kwargs)
            else:
                tmp = LocalFileTarget(is_tmp=self.ext(path, n=0) or True)
                lpath = self.copy(path, tmp.uri(), cache=cache, **kwargs)
            lpath = remove_scheme(lpath)

            def cleanup():
                if not cache and tmp and tmp.exists():
                    tmp.remove()

            f = lpath if yield_path else open(lpath, mode)
            return RemoteFileProxy(f, success_fn=cleanup, failure_fn=cleanup)

        else:  # write or update
            tmp = LocalFileTarget(is_tmp=self.ext(path, n=0) or True)
            lpath = tmp.path

            def cleanup():
                tmp.remove(silent=True)

            def copy_and_cleanup():
                exists = True
                try:
                    exists = tmp.exists()
                    if exists:
                        self.copy(tmp.uri(), path, perm=perm, dir_perm=dir_perm, cache=cache,
                            **kwargs)
                finally:
                    if exists:
                        tmp.remove(silent=True)

            f = lpath if yield_path else open(lpath, mode)
            return RemoteFileProxy(f, success_fn=copy_and_cleanup, failure_fn=cleanup)


class RemoteTarget(FileSystemTarget):

    fs = None

    def __init__(self, path, fs, **kwargs):
        if not isinstance(fs, RemoteFileSystem):
            raise TypeError("fs must be a {} instance, is {}".format(RemoteFileSystem, fs))

        self.fs = fs

        super(RemoteTarget, self).__init__(path, **kwargs)

    def _parent_args(self):
        args, kwargs = super(RemoteTarget, self)._parent_args()
        args += (self.fs,)
        return args, kwargs

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        if os.path.normpath(str(path)).startswith(".."):
            raise ValueError("path {} forbidden, surpasses file system root".format(path))

        path = self.fs.abspath(path)
        super(RemoteTarget, self.__class__).path.fset(self, path)

    @property
    def abspath(self):
        return self.uri()

    def uri(self, **kwargs):
        return self.fs.uri(self.path, **kwargs)

    def copy_to_local(self, dst=None, **kwargs):
        if dst:
            dst = add_scheme(self.fs.local_fs.abspath(get_path(dst)), "file")
        dst = self.copy_to(dst, **kwargs)
        return remove_scheme(dst)

    def copy_from_local(self, src=None, **kwargs):
        src = add_scheme(self.fs.local_fs.abspath(get_path(src)), "file")
        return self.copy_from(src, **kwargs)

    def move_to_local(self, dst=None, **kwargs):
        if dst:
            dst = add_scheme(self.fs.local_fs.abspath(get_path(dst)), "file")
        dst = self.move_to(dst, **kwargs)
        return remove_scheme(dst)

    def move_from_local(self, src=None, **kwargs):
        src = add_scheme(self.fs.local_fs.abspath(get_path(src)), "file")
        return self.move_from(src, **kwargs)

    def load(self, *args, **kwargs):
        # split kwargs that might be designated for remote files
        remote_kwargs, kwargs = self.fs.split_remote_kwargs(kwargs)

        # forward to the localized representation
        with self.localize(mode="r", **remote_kwargs) as loc:
            return loc.load(*args, **kwargs)

    def dump(self, *args, **kwargs):
        # split kwargs that might be designated for remote files
        remote_kwargs, kwargs = self.fs.split_remote_kwargs(kwargs)

        # forward to the localized representation
        with self.localize(mode="a", **remote_kwargs) as loc:
            return loc.dump(*args, **kwargs)


class RemoteFileTarget(FileSystemFileTarget, RemoteTarget):

    @property
    def cache_path(self):
        if not self.fs.cache:
            return None

        return self.fs.cache.cache_path(self.path)

    @contextmanager
    def localize(self, mode="r", perm=None, dir_perm=None, tmp_dir=None, **kwargs):
        if mode not in ["r", "w", "a"]:
            raise Exception("unknown mode '{}', use 'r', 'w' or 'a'".format(mode))

        logger.debug("localizing {!r} with mode '{}'".format(self, mode))

        if mode == "r":
            with self.fs.open(self.path, "r", _yield_path=True, perm=perm, **kwargs) as lpath:
                yield LocalFileTarget(lpath)

        else:  # mode "w" or "a"
            tmp = LocalFileTarget(is_tmp=self.ext(n=1) or True, tmp_dir=tmp_dir)

            # copy to local in append mode
            if mode == "a" and self.exists():
                self.copy_to_local(tmp, **kwargs)

            try:
                yield tmp

                if tmp.exists():
                    self.copy_from_local(tmp, perm=perm, dir_perm=dir_perm, **kwargs)
                else:
                    logger.warning("cannot move non-existing localized target to actual "
                        "representation {!r}".format(self))
            finally:
                tmp.remove()


class RemoteDirectoryTarget(FileSystemDirectoryTarget, RemoteTarget):

    def _child_args(self, path, type):
        args, kwargs = super(RemoteDirectoryTarget, self)._child_args(path, type)
        args += (self.fs,)
        return args, kwargs

    @contextmanager
    def localize(self, mode="r", perm=None, dir_perm=None, tmp_dir=None, **kwargs):
        if mode not in ["r", "w", "a"]:
            raise Exception("unknown mode '{}', use 'r', 'w' or 'a'".format(mode))

        logger.debug("localizing {!r} with mode '{}'".format(self, mode))

        if mode == "r":
            # create a temporary directory
            tmp = LocalDirectoryTarget(is_tmp=True, tmp_dir=tmp_dir)

            # copy contents
            self.copy_to_local(tmp, **kwargs)

            # yield the copy
            try:
                yield tmp
            finally:
                tmp.remove(silent=True)

        else:  # mode "w" or "a"
            # create a temporary directory
            tmp = LocalDirectoryTarget(is_tmp=True, tmp_dir=tmp_dir)

            # copy in append mode, otherwise ensure that it exists
            if mode == "a" and self.exists():
                self.copy_to_local(tmp)
            else:
                tmp.touch()

            # yield the copy
            try:
                yield tmp

                # move back again, first removing current content
                # TODO: keep track of changed contents in "a" mode and copy only those?
                if tmp.exists():
                    self.remove(**kwargs)
                    self.copy_from_local(tmp, perm=perm, dir_perm=dir_perm, **kwargs)
                else:
                    logger.warning("cannot move non-existing localized target to actual "
                        "representation {!r}, leaving original contents unchanged".format(self))
            finally:
                tmp.remove()


RemoteTarget.file_class = RemoteFileTarget
RemoteTarget.directory_class = RemoteDirectoryTarget


class RemoteFileProxy(object):

    def __init__(self, f, close_fn=None, success_fn=None, failure_fn=None):
        super(RemoteFileProxy, self).__init__()

        self.f = f
        self.is_file = not isinstance(f, six.string_types)

        self.close_fn = close_fn
        self.success_fn = success_fn
        self.failure_fn = failure_fn

    def __call__(self):
        return self.f

    def __getattr__(self, attr):
        return getattr(self.f, attr)

    def __enter__(self):
        return self.f.__enter__() if self.is_file else self.f

    def __exit__(self, exc_type, exc_value, traceback):
        # when an exception was raised, the context was not successful
        success = exc_type is None

        # invoke the exit of the file object
        # when its return value is True, it overwrites the success flag
        if getattr(self.f, "__exit__", None) is not None:
            exit_ret = self.f.__exit__(exc_type, exc_value, traceback)
            if exit_ret is True:
                success = True

        if success:
            if callable(self.success_fn):
                self.success_fn()
        else:
            if callable(self.failure_fn):
                self.failure_fn()

        return success

    def close(self, *args, **kwargs):
        ret = self.f.close(*args, **kwargs)

        if callable(self.close_fn):
            self.close_fn()

        return ret



# coding: utf-8
# flake8: noqa

"""
LSF contrib functionality.
"""

__all__ = [
    "get_lsf_version",
    "LSFJobManager", "LSFJobFileFactory",
    "LSFWorkflow",
]


# provisioning imports
from law.contrib.lsf.util import get_lsf_version
from law.contrib.lsf.job import LSFJobManager, LSFJobFileFactory
from law.contrib.lsf.workflow import LSFWorkflow



# coding: utf-8

"""
Law logging setup.
"""

__all__ = [
    "get_logger", "setup_logging", "setup_logger", "create_stream_handler", "is_tty_handler",
    "get_tty_handlers", "Logger", "LogFormatter",
]


from collections import defaultdict
import logging

import six

from law.util import no_value, colored, ipykernel, ON_COLAB


_logging_setup = False


def setup_logging():
    """
    Sets up the internal law loggers as well as all other loggers listed in the ``"logging"`` config
    section as (*name*, *level*) pairs. This includes loggers that do not use the ``"law.*"``
    namespace which can be seen as a convenient feature to set up custom loggers.
    """
    global _logging_setup

    # make sure logging is setup only once
    if _logging_setup:
        return
    _logging_setup = True

    # remove root handlers on colab
    if ON_COLAB and logging.root.handlers:
        logging.root.removeHandler(logging.root.handlers[0])

    # setup the main law logger first and set its handler which is propagated to subloggers
    logger = get_logger("law", skip_setup=True)
    logger = setup_logger(logger, add_console_handler=False)
    logger.addHandler(create_stream_handler())

    # set levels for all loggers and add the console handler for all non-law loggers
    from law.config import Config
    for name, level in Config.instance().items("logging"):
        setup_logger(name, level)


def _logger_setup(logger, value=None):
    attr = "_law_logger_setup"
    if value is not None:
        setattr(logger, attr, value)
    return getattr(logger, attr, False)


def get_logger(*args, **kwargs):
    """ get_logger(*args, skip_setup=False, **kwargs)
    Replacement for *logging.getLogger* that makes sure that the custom :py:class:`Logger` class is
    used when new loggers are created and that the logger is properly set up by
    :py:meth:`setup_logger`.
    """
    skip_setup = kwargs.pop("skip_setup", False)
    orig_cls = logging.getLoggerClass()
    logging.setLoggerClass(Logger)
    try:
        logger = logging.getLogger(*args, **kwargs)

        # set it up once
        if not skip_setup:
            setup_logger(logger)

        return logger
    finally:
        logging.setLoggerClass(orig_cls)


def setup_logger(logger, level=None, add_console_handler=None, clear=False, force=False):
    """
    Sets up a *logger*, optionally given by its name, configures it to have a certain *level* and
    adds a preconfigured console handler when *add_console_handler* is *True*. When
    *add_console_handler* is a dictionary, its items are forwarded as keyword arguments to the
    :py:func:`create_stream_handler` which handles the handler setup internally. When *None*,
    *add_console_handler* is default to *True* in case the logger is not a "law" sublogger and has
    no tty handlers registered yet.

    Each logger is setup only once unless *force* is *True.

    *level* can either be an integer or the name of a level present in the *logging* module. When no
    *level* is  given, the level of the ``"law"`` base logger is used as a default. When the logger
    already existed and *clear* is *True*, all handlers and filters are removed first. The logger
    object is returned.
    """
    # get the logger
    logger = logger if isinstance(logger, logging.Logger) else get_logger(logger, skip_setup=True)
    name = logger.name

    # do nothing when the logger was already set up or force is defined
    if _logger_setup(logger) or force:
        return logger
    _logger_setup(logger, True)

    # sanitize the level
    if level is None:
        from law.config import Config
        level = Config.instance().get_expanded("logging", name, None)
    if isinstance(level, six.string_types):
        level = getattr(logging, level.upper(), None)
    if level is None:
        level = get_logger("law").level

    # clear handlers and filters
    is_existing = name in logging.root.manager.loggerDict
    if is_existing and clear:
        for h in list(logger.handlers):
            logger.removeHandler(h)
        for f in list(logger.filters):
            logger.removeFilter(f)

    # set the level
    logger.setLevel(level)

    # add a console handler
    if add_console_handler is None:
        add_console_handler = not name.startswith("law.") and not get_tty_handlers(name)
    if add_console_handler or isinstance(add_console_handler, dict):
        kwargs = add_console_handler if isinstance(add_console_handler, dict) else {}
        logger.addHandler(create_stream_handler(**kwargs))

    return logger


def create_stream_handler(handler_kwargs=None, formatter_kwargs=None, formatter_cls=no_value):
    """ create_stream_handler(handler_kwargs=None, formatter_kwargs=None, formatter_cls=LogFormatter)
    Creates a new StreamHandler instance, passing all *handler_kwargs* to its constructor, and
    returns it. When not *None*, an instance of *formatter_cls* is created using *formatter_kwargs*
    and added to the handler instance.
    """
    # create the handler
    handler = logging.StreamHandler(**(handler_kwargs or {}))

    # add a formatter
    if formatter_cls == no_value:
        formatter_cls = LogFormatter
    if formatter_cls is not None:
        formatter = formatter_cls(**(formatter_kwargs or {}))
        handler.setFormatter(formatter)

    return handler


def is_tty_handler(handler):
    """
    Returns *True* if a logging *handler* is a *StreamHandler* which logs to a tty (i.e. *stdout* or
    *stderr*), an IPython *OutStream*, or a base *Handler* with a *console* attribute evaluating to
    *True*. The latter check is intended to cover a variety of handlers provided by custom modules.
    """
    if isinstance(handler, logging.StreamHandler) and getattr(handler, "stream", None):
        if callable(getattr(handler.stream, "isatty", None)) and handler.stream.isatty():
            return True
        elif ipykernel and isinstance(handler.stream, ipykernel.iostream.OutStream):
            return True
    if isinstance(handler, logging.Handler) and getattr(handler, "console", None):
        return True
    return False


def get_tty_handlers(logger):
    """
    Returns a list of all handlers of a *logger* that log to a tty.
    """
    if isinstance(logger, six.string_types):
        logger = get_logger(logger)
    return [handler for handler in getattr(logger, "handlers", []) if is_tty_handler(handler)]


class Logger(logging.Logger):
    """
    Custom logger class that adds an additional set of log methods, i.e., :py:meth:`debug_once`,
    :py:meth:`info_once`, :py:meth:`warning_once`, :py:meth:`error_once`, :py:meth:`critical_once`
    and :py:meth:`fatal_once`, that log certain messages only once depending on a string identifier.
    """

    def __init__(self, *args, **kwargs):
        super(Logger, self).__init__(*args, **kwargs)

        # names of logs per level that are issued only once
        self._once_logs = defaultdict(set)

    def debug_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["debug"]:
            self._once_logs["debug"].add(log_id)
            self.debug(*args, **kwargs)

    def info_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["info"]:
            self._once_logs["info"].add(log_id)
            self.info(*args, **kwargs)

    def warning_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["warning"]:
            self._once_logs["warning"].add(log_id)
            self.warning(*args, **kwargs)

    def error_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["error"]:
            self._once_logs["error"].add(log_id)
            self.error(*args, **kwargs)

    def critical_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["critical"]:
            self._once_logs["critical"].add(log_id)
            self.critical(*args, **kwargs)

    def fatal_once(self, log_id, *args, **kwargs):
        # when no log_id is set, but just a message, it is received as log_id
        if not args:
            args = (log_id,)
        if log_id not in self._once_logs["fatal"]:
            self._once_logs["fatal"].add(log_id)
            self.fatal(*args, **kwargs)


class LogFormatter(logging.Formatter):
    """ __init__(*args, log_template=None, err_template=None, level_styles=None, name_styles=None, \
        msg_styles=None, format_level=None, format_name=None, format_msg=None, **kwargs)
    Configurable formatter class for colored logs. When set, *log_template*, *err_template*,
    *level_styles*, *name_styles*, *msg_styles*, *format_level*, *format_name* and *format_msg*
    control the log formats and styles on instance level. When *None*, they default to the upper
    case class level attributes described below. All *args* and *kwargs* are forwarded to
    :py:class:`logging.Formatter`.

    .. py:classattribute:: LOG_TEMPLATE

        type: string

        Template for log messages without stack traces.

    .. py:classattribute:: ERR_TEMPLATE

        type: string

        Template for log messages including stack traces.

    .. py:classattribute:: LEVEL_STYLES

        type: dict

        Style attributes forwarded to :py:func:`law.util.colored` per log level for styling level
        names in logs

    .. py:classattribute:: NAME_STYLES

        type: dict

        Style attributes forwarded to :py:func:`law.util.colored` per log level for styling logger
        names in logs.

    .. py:classattribute:: MSG_STYLES

        type: dict

        Style attributes forwarded to :py:func:`law.util.colored` per log level for styling messages
        in logs.

    .. py:classattribute:: FORMAT_LEVEL

        type: callable or None

        Custom callback to format the log level using the full record.

    .. py:classattribute:: FORMAT_NAME

        type: callable, None

        Custom callback to format the loger name using the full record.

    .. py:classattribute:: FORMAT_MSG

        type: callable, None

        Custom callback to format the log message using the full record.
    """

    LOG_TEMPLATE = "{level}: {name} - {msg}"
    ERR_TEMPLATE = "{level}: {name} - {msg}\n{traceback}"

    LEVEL_STYLES = {
        "DEBUG": {"color": "cyan"},
        "INFO": {"color": "green"},
        "WARNING": {"color": "yellow"},
        "ERROR": {"color": "red"},
        "CRITICAL": {"color": "red", "style": "bright"},
        "FATAL": {"color": "red", "style": "bright"},
    }
    NAME_STYLES = {}
    MSG_STYLES = {
        "WARNING": {"color": "yellow"},
        "ERROR": {"color": "red"},
        "CRITICAL": {"color": "red", "style": "bright"},
        "FATAL": {"color": "red", "style": "bright"},
    }

    FORMAT_LEVEL = None
    FORMAT_NAME = None
    FORMAT_MSG = None

    def __init__(self, *args, **kwargs):
        self.log_template = kwargs.pop("log_template", self.LOG_TEMPLATE)
        self.err_template = kwargs.pop("err_template", self.ERR_TEMPLATE)
        self.level_styles = kwargs.pop("level_styles", self.LEVEL_STYLES)
        self.name_styles = kwargs.pop("name_styles", self.NAME_STYLES)
        self.msg_styles = kwargs.pop("msg_styles", self.MSG_STYLES)
        self.format_level = kwargs.pop("format_level", self.FORMAT_LEVEL)
        self.format_name = kwargs.pop("format_name", self.FORMAT_NAME)
        self.format_msg = kwargs.pop("format_msg", self.FORMAT_MSG)

        super(LogFormatter, self).__init__(*args, **kwargs)

    def format(self, record):
        """"""
        # get and style the level
        level = self.format_level(record) if callable(self.format_level) else record.levelname
        level = colored(level, **self.level_styles.get(record.levelname, {}))

        # get and style the name
        name = self.format_name(record) if callable(self.format_name) else record.name
        name = colored(name, **self.name_styles.get(record.levelname, {}))

        # get and style the message
        msg = self.format_msg(record) if callable(self.format_msg) else record.getMessage()
        msg = colored(msg, **self.msg_styles.get(record.levelname, {}))

        # build template data
        tmpl = self.log_template
        data = dict(level=level, name=name, msg=msg)

        # add traceback and change the template when the record contains exception info
        if record.exc_info:
            tmpl = self.err_template
            data["traceback"] = self.formatException(record.exc_info)

        return tmpl.format(**data)



# coding: utf-8

"""
Keras target formatters.
"""

__all__ = ["KerasModelFormatter", "KerasWeightsFormatter", "TFKerasModelFormatter"]


from law.target.formatter import Formatter
from law.target.file import get_path
from law.logger import get_logger
from law.util import no_value


logger = get_logger(__name__)


class KerasModelFormatter(Formatter):

    name = "keras_model"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith((".hdf5", ".h5", ".json", ".yaml", ".yml"))

    @classmethod
    def load(cls, path, *args, **kwargs):
        import keras

        path = get_path(path)

        # the method for loading the model depends on the file extension
        if path.endswith(".json"):
            with open(path, "r") as f:
                return keras.models.model_from_json(f.read(), *args, **kwargs)

        if path.endswith((".yml", ".yaml")):
            with open(path, "r") as f:
                return keras.models.model_from_yaml(f.read(), *args, **kwargs)

        # .hdf5, .h5, bundle
        return keras.models.load_model(path, *args, **kwargs)

    @classmethod
    def dump(cls, path, model, *args, **kwargs):
        _path = get_path(path)
        perm = kwargs.pop("perm", no_value)

        # the method for saving the model depends on the file extension
        ret = None
        if _path.endswith(".json"):
            with open(_path, "w") as f:
                f.write(model.to_json(*args, **kwargs))

        elif _path.endswith((".yml", ".yaml")):
            with open(_path, "w") as f:
                f.write(model.to_yaml(*args, **kwargs))

        else:  # .hdf5, .h5, bundle
            ret = model.save(_path, *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret


class KerasWeightsFormatter(Formatter):

    name = "keras_weights"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith((".hdf5", ".h5"))

    @classmethod
    def load(cls, path, model, *args, **kwargs):
        return model.load_weights(get_path(path), *args, **kwargs)

    @classmethod
    def dump(cls, path, model, *args, **kwargs):
        perm = kwargs.pop("perm", no_value)

        ret = model.save_weights(get_path(path), *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret


class TFKerasModelFormatter(Formatter):

    name = "tf_keras"

    @classmethod
    def accepts(cls, path, mode):
        return False

    @classmethod
    def load(cls, path, *args, **kwargs):
        # deprecation warning until v0.1
        logger.warning_once("law.contrib.keras.TFKerasModelFormatter is deprecated, please use "
            "law.contrib.tensorflow.TFKerasModelFormatter (named 'tf_keras_model') instead")

        import tensorflow as tf
        return tf.keras.models.load_model(get_path(path), *args, **kwargs)

    @classmethod
    def dump(cls, path, model, *args, **kwargs):
        # deprecation warning until v0.1
        logger.warning_once("law.contrib.keras.TFKerasModelFormatter is deprecated, please use "
            "law.contrib.tensorflow.TFKerasModelFormatter (named 'tf_keras_model') instead")

        perm = kwargs.pop("perm", no_value)

        model.save(get_path(path), *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)



# coding: utf-8

"""
Decorators for task methods for convenient working with ARC.
"""

__all__ = ["ensure_arcproxy"]


from law.decorator import factory
from law.contrib.arc import check_arcproxy_validity


@factory(accept_generator=True)
def ensure_arcproxy(fn, opts, task, *args, **kwargs):
    """
    Decorator for law task methods that checks the validity of the arc proxy and throws an
    exception in case it is invalid. This can prevent late errors on remote worker notes that except
    arc proxies to be present. Accepts generator functions.
    """
    def before_call():
        # check the proxy validity
        if not check_arcproxy_validity():
            raise Exception("arc proxy not valid")

        return None

    def call(state):
        return fn(task, *args, **kwargs)

    def after_call(state):
        return

    return before_call, call, after_call



# coding: utf-8

"""
Helpful decorators to use with tasks.

Example usage:

.. code-block:: python

    class MyTask(law.Task):

        @log
        @safe_output(skip=KeyboardInterrupt)
        def run(self):
            ...

The usage of a decorator without invocation (e.g. ``@log``) is equivalent to the one *with*
invocation (``@log()``), for law to distuinguish between the two cases **always** use keyword
arguments when configuring decorators. Default arguments are applied in either case.
"""

__all__ = [
    "factory", "log", "safe_output", "delay", "notify", "timeit", "localize", "require_sandbox",
]


import sys
import time
import traceback
import functools
import inspect
import random
import socket
import collections
import uuid

import luigi
import six

from law.task.proxy import ProxyTask
from law.sandbox.base import SandboxTask
from law.parameter import get_param, NotifyParameter
from law.target.file import localize_file_targets
from law.target.local import LocalFileTarget
from law.util import (
    no_value, uncolored, make_list, multi_match, human_duration, open_compat, join_generators,
    TeeStream, perf_counter, empty_context,
)
from law.logger import get_logger


logger = get_logger(__name__)


class _CompleteTask(luigi.Task):

    def complete(self):
        return True

    def output(self):
        return None


def factory(**default_opts):
    """
    Factory function to create decorators for tasks' run methods. Default options for the decorator
    function can be given in *default_opts*. The returned decorator can be used with or without
    function invocation. Example:

    .. code-block:: python

        @factory(digits=2)
        def runtime(fn, opts, task, *args, **kwargs):
            t0 = time.time()
            try:
                return fn(task, *args, **kwargs)
            finally:
                t1 = time.time()
                diff = round(t1 - t0, opts["digits"])
                print("runtime: {}".format(diff))

        ...

        class MyTask(law.Task):

            @runtime
            def run(self):
                ...

            # or
            @runtime(digits=3):
            def run(self):
                ...

    In most cases, the created decorators are used to decorate run methods. As intended by luigi,
    run methods can become generators by yielding tasks to declare `dynamic dependencies
    <https://luigi.readthedocs.io/en/stable/tasks.html#dynamic-dependencies>`__. As luigi will
    resume the run method from scratch everytime a new, incomplete dependency is yielded,
    decorators are required to be idempotent. Therefore, a plain definition as shown in the example
    above is not sufficient. A decorator that accepts generator functions should look like the
    following:

    .. code-block:: python

        @factory(digits=2, accept_generator=True)
        def runtime(fn, opts, task, *args, **kwargs):
            def before_call():
                t0 = time.time()
                return t0

            def call(t0):
                return fn(task, *args, **kwargs)

            def after_call(t0):
                t1 = time.time()
                diff = round(t1 - t0, opts["digits"])
                print("runtime: {}".format(diff))

            # optional:
            def on_error(e, t0):
                # called when an exception occured in call,
                # return True to prevent the error from being raised
                ...

            return before_call, call, after_call[, on_error]

    ``before_call()`` is invoked only once. It can be used to setup objects, etc, and its return
    value is passed as a single argument to both ``call()`` and ``after_call()``, even when *None*.
    The former function should (at least) call the actual wrapped function and return its result
    while the latter is intended to execute custom logic afterwards. The use of a 4th function for
    handling exceptions is optional. It is called when an exception is raised inside ``call()`` with
    the exception instance and the return value of ``before_call`` as arguments. When its return
    value is *True*, the error is not raised and the return value of the wrapped function becomes
    *None*.

    A decorator that accepts generator functions can also be used to decorate plain, non-generator
    functions, but not vice-versa. Decorated functions can be called with a keyword argument
    ``skip_decorators`` set to *True* to directly call the originally wrapped function without the
    stack of decorators.
    """
    def wrapper(decorator):
        @functools.wraps(decorator)
        def wrapper(fn=None, **opts):
            _opts = default_opts.copy()
            _opts.update(opts)

            def wrapper(fn):
                # get some default options
                accept_generator = _opts.setdefault("accept_generator", False)
                decorate_run = _opts.setdefault("decorate_run", None)

                # get the originally wrapper function
                # the attribute exists when fn is already a wrapper created by another decorator
                orig_attr = "__law_decorator_original_fn"
                orig_fn = getattr(fn, orig_attr, no_value)
                if orig_fn == no_value:
                    orig_fn = fn

                # when the orignal, wrapped function is a generator, check if the decorator is
                # configured to handle them, and raise a exception if not
                is_gen = inspect.isgeneratorfunction(orig_fn)
                if is_gen and not accept_generator:
                    raise Exception("decorator {} is not configured to decorate a generator "
                        "function {}".format(decorator, orig_fn))

                # when decorator_run is None, guess the decision based on the name of the wrapped fn
                if decorate_run is None:
                    decorate_run = orig_fn.__name__ == "run"

                # define a unique attribute to store the result of before_call() (see below)
                state_attr = "__law_decorator_{}_{}_before_call_result_{}".format(
                    decorator.__module__.replace(".", "_"), decorator.__name__, uuid.uuid4().hex)

                @functools.wraps(fn)
                def wrapper(*args, **kwargs):
                    # check if the decorator stack is to be skipped entirey
                    if kwargs.pop("skip_decorators", False):
                        # args[0] is the task
                        return fn(*args, **kwargs)

                    if accept_generator:
                        # when generator functions are accepted, the decorator is excepted to return
                        # three callbacks: before_call(), call(state), and after_call(state)
                        # the latter two take the return value of the first one as a single argument
                        callbacks = tuple(decorator(fn, _opts, *args, **kwargs))
                        if len(callbacks) not in (3, 4):
                            raise Exception("decorators accepting generator functions must return "
                                "3 or 4 callbacks, got {}".format(len(callbacks)))

                        # extract the callbacks
                        before_call, call, after_call = callbacks[:3]
                        if len(callbacks) == 4:
                            on_error = callbacks[3]
                        else:
                            def on_error(e, state):
                                return False

                        # when then wrapped function returns a generator, invoke before_call() once
                        # for idempotency, and extend the generator to run after_call() and reset()
                        if is_gen:
                            # wrap after_call() as it is required to be a generator
                            def after_call_gen(state):
                                yield _CompleteTask() if decorate_run else None
                                after_call(state)

                            # reset function
                            def reset():
                                yield _CompleteTask() if decorate_run else None
                                setattr(fn, state_attr, no_value)

                            # call before_call once
                            state = getattr(wrapper, state_attr, no_value)
                            if state == no_value:
                                state = before_call()
                                setattr(wrapper, state_attr, state)

                            # wrap on_error() to include the state
                            def _on_error(error):
                                return on_error(error, state)

                            # join the generators, pass the result of before_call
                            return join_generators(call(state), after_call_gen(state), reset(),
                                on_error=_on_error)

                        else:
                            # although configured to handle it, the wrapped function is not a
                            # generator, so just invoke the callbacks serially and handle errors
                            state = before_call()

                            try:
                                result = call(state)
                            except (Exception, KeyboardInterrupt) as error:
                                if on_error(error, state):
                                    result = None
                                else:
                                    raise

                            after_call(state)

                            return result

                    else:
                        # the wrapped function is a plain callable, so just call it
                        return decorator(fn, _opts, *args, **kwargs)

                # store the originally wrapped function as an attribute of the wrapper
                setattr(wrapper, orig_attr, orig_fn)

                return wrapper
            return wrapper if fn is None else wrapper(fn)
        return wrapper
    return wrapper


def get_task(task):
    return task if not isinstance(task, ProxyTask) else task.task


@factory(accept_generator=False)
def log(fn, opts, task, *args, **kwargs):
    """ log()
    Wraps a bound method of a task and redirects output of both stdout and stderr to the file
    defined by the tasks's *log_file* parameter or *default_log_file* attribute. If its value is
    ``"-"`` or *None*, the output is not redirected. Does **not** accept generator functions.
    """
    _task = get_task(task)
    log = get_param(_task.log_file, _task.default_log_file)
    if log and not isinstance(log, LocalFileTarget):
        log = str(log)

    if log == "-" or not log:
        return fn(task, *args, **kwargs)
    else:
        # use the local target functionality to create the parent directory
        LocalFileTarget(log).parent.touch()
        with open_compat(log, "a", 1) as f:
            tee = TeeStream(f, sys.__stdout__)
            sys.stdout = tee
            sys.stderr = tee
            try:
                ret = fn(task, *args, **kwargs)
            except:
                traceback.print_exc(file=tee)
                raise
            finally:
                sys.stdout = sys.__stdout__
                sys.stderr = sys.__stderr__
                tee.flush()
        return ret


@factory(skip=None, accept_generator=True)
def safe_output(fn, opts, task, *args, **kwargs):
    """ safe_output(skip=None)
    Wraps a bound method of a task and guards its execution. If an exception occurs, and it is not
    an instance of *skip*, the task's output is removed prior to the actual raising. Accepts
    generator functions.
    """
    def before_call():
        return None

    def call(state):
        return fn(task, *args, **kwargs)

    def after_call(state):
        return

    def on_error(error, state):
        if opts["skip"] is None or not isinstance(error, opts["skip"]):
            for outp in luigi.task.flatten(task.output()):
                outp.remove()

    return before_call, call, after_call, on_error


@factory(t=5.0, stddev=0.0, pdf="gauss", accept_generator=True)
def delay(fn, opts, task, *args, **kwargs):
    """ delay(t=5.0, stddev=0.0, pdf="gauss")
    Wraps a bound method of a task and delays its execution by *t* seconds. Accepts generator
    functions.
    """
    def before_call():
        return None

    def call(state):
        if opts["stddev"] <= 0:
            t = opts["t"]
        elif opts["pdf"] == "gauss":
            t = random.gauss(opts["t"], opts["stddev"])
        elif opts["pdf"] == "uniform":
            t = random.uniform(opts["t"], opts["stddev"])
        else:
            raise ValueError("unknown delay decorator pdf '{}'".format(opts["pdf"]))

        time.sleep(max(t, 0))

        return fn(task, *args, **kwargs)

    def after_call(state):
        return

    return before_call, call, after_call


@factory(on_success=True, on_failure=True, accept_generator=True)
def notify(fn, opts, task, *args, **kwargs):
    """ notify(on_success=True, on_failure=True)
    Wraps a bound method of a task and guards its execution. Information about the execution (task
    name, duration, etc) is collected and dispatched to all notification transports registered on
    wrapped task via adding :py:class:`law.NotifyParameter` parameters. Example:

    .. code-block:: python

        class MyTask(law.Task):

            notify_mail = law.NotifyMailParameter()

            @notify
            # or
            @notify(sender="foo@bar.com", recipient="user@host.tld")
            def run(self):
                ...

    When the *notify_mail* parameter is *True*, a notification is sent to the configured email
    address. Also see :ref:`config-notifications`. Accepts generator functions.
    """
    _task = get_task(task)

    def before_call():
        # prepare notification transports
        transports = []
        for param_name, param in _task.get_params():
            if isinstance(param, NotifyParameter) and getattr(_task, param_name):
                try:
                    transport = param.get_transport()
                    if transport:
                        transports += make_list(transport)
                except Exception as e:
                    logger.warning("get_transport() failed for '{}' parameter: {}".format(
                        param_name, e))

        # get a timestamp
        t0 = perf_counter()

        return transports, t0

    def call(state):
        return fn(task, *args, **kwargs)

    def send(error, transports, t0):
        # do nothing when there are no transports
        if not transports:
            return

        # do nothing on KeyboardInterrupt, or when on_success / on_failure do not match the status
        success = error is None
        if isinstance(error, KeyboardInterrupt):
            return
        elif success and not opts["on_success"]:
            return
        elif not success and not opts["on_failure"]:
            return

        # prepare message content
        duration = human_duration(seconds=round(perf_counter() - t0, 1))
        status_string = "succeeded" if success else "failed"
        title = "Task {} {}!".format(_task.get_task_family(), status_string)
        parts = collections.OrderedDict([
            ("Host", socket.gethostname()),
            ("Duration", duration),
            ("Last message", "-" if not len(_task._message_cache) else _task._message_cache[-1]),
            ("Task", str(_task)),
        ])
        if not success:
            parts["Traceback"] = traceback.format_exc()
        message = "\n".join("{}: {}".format(*tpl) for tpl in parts.items())

        # dispatch via all transports
        for transport in transports:
            fn = transport["func"]
            raw = transport.get("raw", False)
            colored = transport.get("colored", False)

            # remove color commands if necessary
            if not colored:
                _title = uncolored(title)
                if raw:
                    _content = {
                        k: (uncolored(v) if isinstance(v, six.string_types) else v)
                        for k, v in parts.items()
                    }
                else:
                    _content = uncolored(message)
            else:
                _title = title
                _content = parts.copy() if raw else message

            # invoke the function
            try:
                fn(success, _title, _content, **opts)
            except Exception as e:
                t = traceback.format_exc()
                logger.warning("notification via transport '{}' failed: {}\n{}".format(fn, e, t))

    def after_call(state):
        return send(None, *state)

    def on_error(error, state):
        return send(error, *state)

    return before_call, call, after_call, on_error


@factory(accept_generator=True)
def timeit(fn, opts, task, *args, **kwargs):
    """ timeit()
    Wraps a bound method of a task and logs its execution time in a human readable format using the
    task's logger instance in info mode. Accepts generator functions.
    """
    def before_call():
        t0 = perf_counter()
        return t0

    def call(t0):
        return fn(task, *args, **kwargs)

    def log_duration(t0):
        duration = human_duration(seconds=round(perf_counter() - t0, 1))
        task.logger.info("runtime: {}".format(duration))

    def after_call(t0):
        log_duration(t0)

    def on_error(error, t0):
        log_duration(t0)

    return before_call, call, after_call, on_error


@factory(input=True, output=True, input_kwargs=None, output_kwargs=None, accept_generator=False)
def localize(fn, opts, task, *args, **kwargs):
    """ localize(input=True, output=True, input_kwargs=None, output_kwargs=None)
    Wraps a bound method of a task and temporarily changes the input and output methods to return
    localized targets. When *input* (*output*) is *True*, :py:meth:`Task.input`
    (:py:meth:`Task.output`) is adjusted. *input_kwargs* and *output_kwargs* can be dictionaries
    that are passed as keyword arguments to the respective localization method. Does **not** accept
    generator functions.
    """
    # store original input and output methods
    input_orig = None
    output_orig = None
    if opts["input"]:
        input_orig = (
            task.__getattribute__("input", proxy=False)
            if isinstance(task, SandboxTask)
            else task.input
        )
    if opts["output"]:
        output_orig = (
            task.__getattribute__("output", proxy=False)
            if isinstance(task, SandboxTask)
            else task.output
        )

    # wrap input context
    input_context = empty_context
    if opts["input"]:
        def input_context():  # noqa: F811
            input_struct = task.input()
            input_kwargs = opts["input_kwargs"] or {}
            input_kwargs.setdefault("mode", "r")
            return localize_file_targets(input_struct, **input_kwargs)

    # wrap output context
    output_context = empty_context
    if opts["output"]:
        def output_context():  # noqa: F811
            output_struct = task.output()
            output_kwargs = opts["output_kwargs"] or {}
            output_kwargs.setdefault("mode", "w")
            return localize_file_targets(output_struct, **output_kwargs)

    try:
        # localize both target contexts
        with input_context() as localized_inputs, output_context() as localized_outputs:
            # patch the input method to always return the localized inputs
            if opts["input"]:
                def input_patched(self):
                    return localized_inputs

                task.input = _patch_localized_method(task, input_patched)

            # patch the output method to always return the localized outputs
            if opts["output"]:
                def output_patched(self):
                    return localized_outputs

                task.output = _patch_localized_method(task, output_patched)

            return fn(task, *args, **kwargs)

    finally:
        # restore the methods
        if input_orig is not None:
            task.input = input_orig
        if output_orig is not None:
            task.output = output_orig


def _patch_localized_method(task, func):
    # add a flag to func
    func._patched_localized_method = True

    # bind to task
    return func.__get__(task)


def _is_patched_localized_method(func):
    return getattr(func, "_patched_localized_method", False) is True


@factory(sandbox=None, accept_generator=True)
def require_sandbox(fn, opts, task, *args, **kwargs):
    """ require_sandbox(sandbox=None)
    Wraps a bound method of a sandbox task and throws an exception when the method is called while
    the task is not sandboxed yet. This is intended to prevent undesired results or non-verbose
    error messages when the method is invoked outside the requested sandbox. When *sandbox* is set,
    it can be a (list of) pattern(s) to compare against the task's effective sandbox and in error is
    raised if they don't match. Accepts generator functions.
    """
    def before_call():
        if not isinstance(task, SandboxTask):
            raise TypeError("require_sandbox can only be used to decorate methods of tasks that "
                "inherit from SandboxTask, got '{!r}'".format(task))

        if not task.is_sandboxed():
            raise Exception("the invocation of method {} requires task {!r} to be sandboxed".format(
                fn.__name__, task))

        if opts["sandbox"] and not multi_match(task.effective_sandbox, make_list(opts["sandbox"])):
            raise Exception("the invocation of method {} requires the sandbox of task {!r} to "
                "match '{}'" .format(fn.__name__, task, opts["sandbox"]))

        return None

    def call(state):
        return fn(task, *args, **kwargs)

    def after_call(state):
        return

    return before_call, call, after_call



# coding: utf-8

"""
LSF utilities.
"""

__all__ = ["get_lsf_version"]


import re
import subprocess
import threading

from law.util import no_value, interruptable_popen


_lsf_version = no_value
_lsf_version_lock = threading.Lock()


def get_lsf_version():
    """
    Returns the version of the LSF installation in a 3-tuple. The value is cached to accelerate
    repeated function invocations.
    """
    global _lsf_version

    if _lsf_version == no_value:
        version = None
        with _lsf_version_lock:
            code, out, _ = interruptable_popen("bjobs -V", shell=True, executable="/bin/bash",
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            if code == 0:
                first_line = out.strip().split("\n")[0].strip()
                m = re.match(r"^Platform LSF (\d+)\.(\d+)\.(\d+).+$", first_line)
                if m:
                    version = tuple(map(int, m.groups()))

            _lsf_version = version

    return _lsf_version
